generate_data <- function(
  n_sigmasq = 1,
  n_phi     = 10,
  n_tausq   = 0 ,
  str_cov_model = "exponential",
  n_kappa   = 0.5,
  n_of_trials = 1, # number of trials per location (can actually also be a vector)
  bo_irreg  = FALSE, # Add some irregularity in locations
  bo_determ = FALSE, # Deterministic rather than stochastic binary layer (m=1)
  n_beta_0   = 0,
  n_beta_1   = 0,    
  df_simulation_field = NULL
)
{  
  

  #### >>>> Generate the simulated "observations" ####
  
  ## if df_simulation_field not provided, generate one 
  if(is.null(df_simulation_field))
  {
    
    ## Basic grid
    df_simulation_field <- expand.grid(
      seq(from = 0, to = 100, by = 3),
      seq(from = 0, to = 100, by = 3)
    )
  }  
  
  if (bo_irreg)
  {
    df_simulation_field[,1] <- jitter(df_simulation_field[,1], factor=1.5)
    df_simulation_field[,2] <- jitter(df_simulation_field[,2], factor=1.5)
  }
  
  
  
  ## Add a basic trend
  vn_covariate <- df_simulation_field[,1]+df_simulation_field[,2]
  vn_mean      <-  n_beta_0 + n_beta_1 * vn_covariate
  
  
  
  ##  Simulate basic field (using package geoR)  
  grf <- grf(n = nrow(df_simulation_field), 
             grid = df_simulation_field[,1:2], 
             nsim = 1, 
             cov.model = str_cov_model,
             cov.pars = c(n_sigmasq, n_phi), 
             nugget = n_tausq, 
             kappa = n_kappa,
             mean   = vn_mean, 
             RF    = TRUE, 
             messages = TRUE)
  
  
  vn_signal <- grf$data
  
  #browser()
  
  ## Extra: deterministic binary layer
  if(bo_determ)
  {
    vn_data <- ifelse(grf$data > 0, 1, 0)
  } else  # Bernoulli
  {
    if (n_of_trials == 1.5)
    {
      vn_of_trials <- c(1,2)
    } else
    {
      vn_of_trials <- n_of_trials
    }
    

    ## Limit range to avoid numerical problems
    vn_signal <- pmin(vn_signal, 500)
    
    vn_data <- rbinom(n = length(vn_signal),      # size of sample
                      size = vn_of_trials,        # trials per location
                      prob = exp(vn_signal)/(1 + exp(vn_signal)))
  }
  
  #df_simulation_field <- cbind(df_simulation_field, li_sim$data, li_sim$latent)
  df_simulation_field <- cbind(df_simulation_field[,1:2], 
                               vn_covariate, 
                               vn_data, 
                               vn_signal, 
                               rep(vn_of_trials,
                                   length.out = nrow(df_simulation_field) 
                                   )
                               )
  
  names(df_simulation_field) <- c("x", "y", "covariate", "data", "signal", "number_of_trials")
  
  geod_cali <- as.geodata(obj        = df_simulation_field,
                          coords.col = 1:2, 
                          data.col   = 4,
                          covar.col  = 3,
                          units.m.col = 6
  )
  
  
  
  geod_cali_signal <- as.geodata(obj    = df_simulation_field,
                          coords.col = 1:2, 
                          data.col   = 5,
                          covar.col  = 3
  )
  

  
  ## Grid for prediction
  df_prediction_field <- expand.grid(
    seq(from = 0, to = 100, by = 3),
    seq(from = 0, to = 100, by = 3)
  )
  
  vn_prediction_covariate = df_prediction_field[,1]+df_prediction_field[,2]
  
  df_prediction_field <- cbind(df_prediction_field,
                               vn_prediction_covariate)
  
  names(df_prediction_field) <- c("x", "y", "covariate")
  
  geod_prediction_field <- as.geodata(obj    = df_simulation_field,
                                 coords.col = 1:2, 
                                 data.col   = 3,
                                 covar.col  = 3
                                )
  
  
  return(list (
    geod_cali        = geod_cali,
    geod_cali_signal = geod_cali_signal,
    geod_prediction_field = geod_prediction_field,
    df_simulation_field = df_simulation_field,
    df_prediction_field = df_prediction_field
    )
  )
}
